
#ifndef PLS_SCHEDULER_H
#define PLS_SCHEDULER_H

#include <array>
#include <iostream>

#include "pls/internal/helpers/profiler.h"

#include "pls/internal/base/aligned_stack.h"
#include "pls/internal/base/thread.h"
#include "pls/internal/base/barrier.h"

#include "thread_state.h"
#include "root_task.h"
#include "scheduler_memory.h"

namespace pls {
    namespace internal {
        namespace scheduling {
            void worker_routine();
            using scheduler_thread = base::thread<decltype(&worker_routine), thread_state>;

            class scheduler {
                friend void worker_routine();

                const unsigned int num_threads_;
                scheduler_memory* memory_;

                base::barrier sync_barrier_;
                bool terminated_;
            public:
                explicit scheduler(scheduler_memory* memory, unsigned int num_threads);
                ~scheduler();

                template<typename Function>
                void perform_work(Function work_section) {
                    PROFILE_WORK_BLOCK("scheduler::perform_work")
                    root_task<Function> master{work_section};

                    // Push root task on stacks
                    auto new_master = memory_->task_stack_for(0)->push(master);
                    memory_->thread_state_for(0)->root_task_ = new_master;
                    memory_->thread_state_for(0)->current_task_ = new_master;
                    for (unsigned int i = 1; i < num_threads_; i++) {
                        root_worker_task<Function> worker{new_master};
                        auto new_worker = memory_->task_stack_for(0)->push(worker);
                        memory_->thread_state_for(i)->root_task_ = new_worker;
                        memory_->thread_state_for(i)->current_task_ = new_worker;
                    }

                    // Perform and wait for work
                    sync_barrier_.wait(); // Trigger threads to wake up
                    sync_barrier_.wait(); // Wait for threads to finish

                    // Clean up stack
                    memory_->task_stack_for(0)->pop<typeof(master)>();
                    for (unsigned int i = 1; i < num_threads_; i++) {
                        root_worker_task<Function> worker{new_master};
                        memory_->task_stack_for(0)->pop<typeof(worker)>();
                    }
                }

                // TODO: See if we should place this differently (only for performance reasons)
                template<typename Task>
                static void execute_task(Task& task, int depth=-1) {
                    static_assert(std::is_base_of<abstract_task, Task>::value, "Only pass abstract_task subclasses!");

                    auto my_state = base::this_thread::state<thread_state>();
                    abstract_task* old_task;
                    abstract_task* new_task;

                    // Init Task
                    {
                        std::lock_guard<base::spin_lock> lock{my_state->lock_};
                        old_task = my_state->current_task_;
                        new_task = my_state->task_stack_->push(task);

                        new_task->set_depth(depth >= 0 ? depth : old_task->depth() + 1);
                        my_state->current_task_ = new_task;
                        old_task->set_child(new_task);
                    }

                    // Run Task
                    new_task->execute();

                    // Teardown state back to before the task was executed
                    {
                        std::lock_guard<base::spin_lock> lock{my_state->lock_};

                        old_task->set_child(nullptr);
                        my_state->current_task_ = old_task;

                        my_state->task_stack_->pop<Task>();
                    }
                }

                static abstract_task* current_task() { return base::this_thread::state<thread_state>()->current_task_; }

                void terminate(bool wait_for_workers=true);
                unsigned int num_threads() const { return num_threads_; }
                thread_state* thread_state_for(size_t id) { return memory_->thread_state_for(id); }
            };
        }
    }
}

#endif //PLS_SCHEDULER_H
